/********** nlistd.c
	    NList daemon that performs background tasks for nlist
 		Written by Gabor Paller
						*********/

#include	<stdio.h>
#include	<stdlib.h>
#include	<errno.h>
#include	<sys/types.h>
#include	<sys/fcntl.h>
#include	<sys/param.h>
#include	<sys/ioctl.h>
#include	<sys/wait.h>
#include	<unistd.h>
#include	<string.h>
#include	<stdarg.h>
#include	<setjmp.h>
#include	<time.h>
#include	<sys/termios.h>
#if defined SOLARIS
#include	<sys/varargs.h>
#endif
#define		N_EXTERN
#define		MAIN
#include	"config.h"
#include	"commonvars.h"

/* Uncomment if you want to compile the unit test */
/* #define DEBUG */

#define ARTICLEID struct articleid
struct articleid {
  char	articleid[256];
  ARTICLEID *link;
};

ARTICLEID	*article_list;

/* Time when the newsgroup was checked last time */
unsigned long	lastcheck_time;
unsigned long	firstcheck_time;
long		last_read_article_id;
long		current_article_id;
long		first_article_id;

/* temporary buffer for NNTP messages */
static char	tempbuffer[1002];
static char	from_addr[N_BUFSIZE];
static bool	nlist_header_found;

/* Reads the status variables of nlistd and substitutes them
   with default values if the status file doesn't exist */
void read_status() {
  FILE	*f;
  char	fname[N_BUFSIZE];
  unsigned long	l;
  
  sprintf( fname,"%s/%s.nlistd_status",PATH,list_name );
  lastcheck_time = (unsigned long)time( NULL );
  firstcheck_time = 0;
  current_article_id = -1;
  last_read_article_id = -1;
  if( ( f = fopen( fname,"r" ) ) == NULL )
  	goto end;
  while( fgets( fname,N_BUFSIZE-1,f ) != NULL ) {
    if( strncmp( fname,"lastcheck_time=",15 ) == 0 ) {
      if( sscanf( fname+15,"%lu",&l ) < 1 )
      	continue;
      lastcheck_time = l;
    }
    if( strncmp( fname,"firstcheck_time=",16 ) == 0 ) {
      if( sscanf( fname+16,"%lu",&l ) < 1 )
      	continue;
      firstcheck_time = l;
    }
    if( strncmp( fname,"current_article_id=",19 ) == 0 ) {
      if( sscanf( fname+19,"%li",&l ) < 1 )
      	continue;
      current_article_id = l;
    }
    if( strncmp( fname,"last_read_article_id=",21 ) == 0 ) {
      if( sscanf( fname+21,"%li",&l ) < 1 )
      	continue;
      last_read_article_id = l;
    }
  }
  fclose( f );
/* Just for paranoia, firstcheck_time must always be present */
end:
  if( firstcheck_time == 0L )
  	firstcheck_time = lastcheck_time;
}

/* Writes the status variables of nlistd */
void write_status() {
  FILE	*f;
  char	fname[N_BUFSIZE];
  
  sprintf( fname,"%s/%s.nlistd_status",PATH,list_name );
  if( ( f = fopen( fname,"w" ) ) == NULL )
  	ehandle( 1,"nlistd: cannot write status file %s",fname );
  fprintf( f,"firstcheck_time=%lu\n",firstcheck_time );
  fprintf( f,"lastcheck_time=%lu\n",lastcheck_time );
  fprintf( f,"current_article_id=%li\n",current_article_id );
  fprintf( f,"last_read_article_id=%li\n",last_read_article_id );
  fclose( f );
}

/* Deallocates the article ID list */
void free_article_list() {
  ARTICLEID	*a;
  
  while( article_list != NULL ) {
    a = article_list;
    article_list = article_list->link;
    free( a );
  }
}

/* Selects nntp_newsgroup group. Returns 0 if OK, error code else.
   NNTP session must already be open */
int select_group() {
  char	*p,*p1;
  
  sprintf( tempbuffer,"GROUP %s\n",nntp_newsgroup );
  socket_write_line( tempbuffer );
  socket_read_line( tempbuffer,1000 );
/* If response code is not 2xx, return with error */
  if( *tempbuffer != '2' ) {
    Log( 1,1,"nlistd: NNTP group selection failed: %s",tempbuffer );
    return 1;
  }
  p = strtok( tempbuffer, " \t" );	/* Status code */
  if( p == NULL )
  	goto l0;
  p = strtok( NULL," \t" );		/* Estimated number of articles */
  if( p == NULL )
  	goto l0;
  p1 = strtok( NULL," \t" );		/* First article number */
  if( p1 == NULL )
  	goto l0;
  p = strtok( NULL," \t" );		/* Last article number */
  if( p == NULL )
  	goto l0;
/* Update current article ID */
  sscanf( p1,"%li",&first_article_id );
  sscanf( p,"%li",&current_article_id );
l0:
  return 0;
}

/* Processes the new news in the newsgroup. Fetches the message ids
   in the last newschecktime period and builds a linked list of article
   IDs pointed to by article_list */
void process_new_news_by_newnews( unsigned long lct ) {
  time_t	t;
  struct tm	*tmp;
  int	year;
  char	*p;
  ARTICLEID	*a;
  
  t = (time_t)lct;
  tmp = gmtime( &t );
  year = tmp->tm_year;
  if( year > 99 )
  	year -= 100;
/* Delete previous article ID list - if any */
  free_article_list();
/* Open NNTP session now to the news server */
  if( open_nntp_session() ) {
    Log( 1,1,"NNTP session could not be opened to %s",nntp_server );
    return;
  }
  sprintf( tempbuffer,"NEWNEWS %s %02u%02u%02u %02u%02u%02u GMT\n",
  				nntp_newsgroup,
  				year, tmp->tm_mon+1, tmp->tm_mday,
  				tmp->tm_hour, tmp->tm_min, tmp->tm_sec );
  socket_write_line( tempbuffer );
  socket_read_line( tempbuffer,1000 );
  if( strncmp( tempbuffer,"230",3 ) != 0 ) {
    Log( 1,1,"nlistd: NNTP new article fetch failed: %s",tempbuffer );
    close_socket();
    return;
  }
  while( *tempbuffer != '.' ) {
    socket_read_line( tempbuffer,1000 );
    if( ( p = strtok( tempbuffer,"\r\n\t " ) ) == NULL )
    	break;
    if( *p != '.' ) {	/* We have an article ID */
      if( !CheckArticleFile( p ) ) {
 /* It was not in the article file - add it to the list of
    articles to be posted to the list */
         a = (ARTICLEID *)malloc( sizeof( ARTICLEID ) );
         strcpy( a->articleid,p );
         a->link = article_list;
         article_list = a;
      }
    }
  }
  close_socket();
}

/* Processes new articles by enumerating article IDs. The NEWNEWS
   method implemented in process_new_news is more efficient but
   NEWNEWS is not available on every news server */
void process_new_news_by_id( unsigned long lct ) {
  long	nl_id,old_id,l,save_l;
  char	*p1,*p2;
  ARTICLEID	*a;
  bool		nextid_found;
  
  if( open_nntp_session() ) {
    Log( 1,1,"NNTP session could not be opened to %s",nntp_server );
    return;
  }
/* Delete previous article ID list - if any */
  free_article_list();
  old_id = current_article_id;
  if( select_group() ) {
    close_socket();
    return;
  }
/* If we didn't read anything from this newsgroup, we find out
   what is the last article. This will be our reference point in
   the future. Now it is assumed that the last article has been
   read because we start relaying only when new articles are posted
   and not the already posted ones */
  if( old_id < 0 ) {
    old_id = current_article_id;
    sprintf( tempbuffer,"LISTGROUP %s\n",nntp_newsgroup );
    socket_write_line( tempbuffer );
    socket_read_line( tempbuffer,1000 );
    if( *tempbuffer != '2' ) {
      close_socket();
      return;
    }
    do {
      socket_read_line( tempbuffer, 1000 );
      if( sscanf( tempbuffer,"%li",&l ) == 1 ) {
        if( l > last_read_article_id )
        	last_read_article_id = l;
      }
    } while( *tempbuffer != '.' );
 /* Strange, the newsgroup can be selected but there are
    no articles listed. Maybe it is empty? */ 
    if( last_read_article_id < 0 ) {
      close_socket();
      return;
    }
    Log( 4,1,"nlistd: starting to relay from article %li, high watermark %li",
    			last_read_article_id,current_article_id );
  }
/* Check if there's anything to download */
  if( old_id >= current_article_id ) {
    close_socket();
    return;
  }
  sprintf( tempbuffer,"STAT %li\n",last_read_article_id );
  socket_write_line( tempbuffer );
  socket_read_line( tempbuffer,1000 );
/* Waiting for 2xx status codes. The article with last_read_article_id number
   may have gone away in rarely used newsgroup so it is not a serious problem
   if the article cannot be STAT'd */
  if( *tempbuffer != '2' ) {
    Log( 4,1,"nlistd: cannot STAT id %li: %s",last_read_article_id,tempbuffer );
/* The last read article has gone away. We have to list the newsgroup to find the
   article whose ID is larger than the last_read_article_id */
    sprintf( tempbuffer,"LISTGROUP %s\n",nntp_newsgroup );
    socket_write_line( tempbuffer );
    socket_read_line( tempbuffer,1000 );
    if( *tempbuffer != '2' ) {
      close_socket();
      return;
    }
    nextid_found = FALSE;
    do {
      socket_read_line( tempbuffer, 1000 );
      if( sscanf( tempbuffer,"%li",&l ) == 1 ) {
        save_l = l;
        if( !nextid_found && ( l > last_read_article_id ) ) {
          last_read_article_id = l;
          nextid_found = TRUE;
        }
      }
    } while( *tempbuffer != '.' );
/* If the next id was not found then we have a problem: our ID
   variables and the newsgroup has desynchronized. We set both current_article_id
   and last_read_article_id to the last message of the newsgroup */
    if( !nextid_found ) {
      current_article_id = save_l;
      last_read_article_id = save_l;
      Log( 3,1,"Newsgroup resynchronization complete, new watermark: %li",save_l );
      close_socket();
      return;
    }
/* Let's try again with the new ID */
    sprintf( tempbuffer,"STAT %li\n",last_read_article_id );
    socket_write_line( tempbuffer );
    socket_read_line( tempbuffer,1000 );
/* Waiting for 2xx status codes. If it isn't, it is hopeless and we give up.
   The news server doesn't even have the articles whose IDs it has just listed */
    if( *tempbuffer != '2' ) {
      Log( 1,1,"Couldn't STAT id %li after resync",last_read_article_id );
      close_socket();
      return;
    }
  }
  while( TRUE ) {
    sprintf( tempbuffer,"NEXT\n" );
    socket_write_line( tempbuffer );
    socket_read_line( tempbuffer,1000 );
    if( *tempbuffer != '2' ) {
      Log( 4,1,"NEXT returned error code: %s",tempbuffer );
      break;
    }
    p1 = strtok( tempbuffer, " \t\n" );
    if( p1 == NULL )
    	goto l0;
/* Article number */
    p1 = strtok( NULL, " \t\n" );
    if( p1 == NULL )
    	goto l0;
    if( sscanf( p1,"%li",&l ) < 1 )
    	goto l0;
/* Article ID */
    p2 = strtok( NULL, " \t\n" );
    if( p2 == NULL )
    	goto l0;
    if( sscanf( p1,"%li",&old_id ) < 1 )
    	goto l0;
#ifdef DEBUG
	printf( "Article number: %li article id: %s\n",old_id,p2 );
#endif
    a = (ARTICLEID *)malloc( sizeof( ARTICLEID ) );
    strcpy( a->articleid,p2 );
    a->link = article_list;
    article_list = a;
/* Hopefully article IDs just get larger but you never know with INND ... */
    if( last_read_article_id < l )
    	last_read_article_id = l;    
  }
l0:
  close_socket();
}

/* Facade function for invoking the appropriate process_new_news_*
   function */
void process_new_news( unsigned long lct ) {
  switch( nntp_newart_access ) {
    case NEWART_STAT:
  	process_new_news_by_id( lct );
  	break;
  	
    case NEWART_NEWNEWS:
  	process_new_news_by_newnews( lct );
  	break;
  }
}

/* Fetches a news article from the news server and writes it
   into the nlistd.tmp file. Returns 0 if OK, error code else */
int fetch_news_article( char message_id[] ) {
  char	fname[N_BUFSIZE];
  char	*p,*p1;
  FILE	*f;
  int	n;
  
  if( open_nntp_session() ) {
    Log( 1,1,"NNTP session could not be opened to %s",nntp_server );
    return 2;
  }
  if( select_group() ) {
    close_socket();
    return 1;
  }
  sprintf( tempbuffer,"ARTICLE %s\n",message_id );
  socket_write_line( tempbuffer );
  socket_read_line( tempbuffer,1000 );
/* Response code must be "headers and article" */
  if( strncmp( tempbuffer,"220",3 ) != 0 ) {
    Log( 1,1,"nlistd: NNTP article fetch [%s] failed: %s",
    			message_id,tempbuffer );
    close_socket();
    return 2;
  }
  sprintf( fname,"%s/%s.nlistd_tmp",PATH,list_name );
  if( ( f = fopen( fname,"w" ) ) == NULL ) {
    Log( 1,1,"nlistd: cannot create file %s",fname );
    close_socket();
    return 3;
  }
  strcpy( from_addr,"" );
  nlist_header_found = FALSE;
  do {
    socket_read_line( tempbuffer,1000 );
    if( ( p = strchr( tempbuffer,(char)0x0D ) ) != NULL )
      	*p = 0;
    if( ( p = strchr( tempbuffer,(char)0x0A ) ) != NULL )
      	*p = 0;    
    if( ( strncmp( tempbuffer,"Path:",5 ) == 0 ) ||
        ( strncmp( tempbuffer,"NNTP-Posting-Host:",18 ) == 0 ) ||
        ( strncmp( tempbuffer,"NNTP-Posting-Date:",18 ) == 0 ) ||
        ( strncmp( tempbuffer,"Newsgroups:",11 ) == 0 ) ||
        ( strncmp( tempbuffer,"Lines:",6 ) == 0 ) ||
        ( strncmp( tempbuffer,"X-Trace:",8 ) == 0 ) ||
        ( strncmp( tempbuffer,"Xref:",5 ) == 0 ) || 
        ( strncmp( tempbuffer,"X-Complaints-To:",16 ) == 0 )
      )
    	continue;
    if( strncmp( tempbuffer,"X-Sender: nlist",15 ) == 0 )
    	nlist_header_found = TRUE;
    if( *tempbuffer != '.' )
		fprintf( f,"%s\n",tempbuffer );
/* We will need the e-mail address (without the sender name)
   for the nlist From header. We extract it here from the From:
   header */ 
    if( strncmp( tempbuffer,"From:",5 ) == 0 ) {
      if( ( p = strchr( tempbuffer+6,'<' ) ) == NULL )
      		strcpy( from_addr,tempbuffer+6 );
      else {
        p1 = strchr( p,'>' );
        if( p1 == NULL )
        	n = strlen( p+1 );
        else
        	n = p1 - p - 1;
        strncpy( from_addr,p+1,n );
        *(from_addr+n) = 0;
      }
    } 
  } while( *tempbuffer != '.' );
  close_socket();
  fclose( f );
  return 0;
}

/* Sends a downloaded news article to nlist */
void send_article_to_list() {
  char	fname[N_BUFSIZE];
  FILE	*f;
  int	fds[2],fc,newpid,numread;
  char	*command_array[3];

/* Don't send out if it comes from nlist */
  if( nlist_header_found )
  	return;
  sprintf( fname,"%s/%s.nlistd_tmp",PATH,list_name );
  if( ( f = fopen( fname,"r" ) ) == NULL ) {
    Log( 1,1,"nlistd: cannot read file %s",fname );
    return;
  }
  command_array[0] = "nlist";
  command_array[1] = list_name;
  command_array[2] = NULL;
  if( pipe( fds ) == -1 ) {
    Log(1,1, "nlistd: cannot create pipe to nlist\n");
    fclose( f );
    return;
  }
  fc = fcntl(fds[1], F_GETFL, 0);
  fcntl(fds[1], F_SETFL, fc & ~O_NONBLOCK);
  newpid = fork();
/* Parent process passes the article to nlist */
  if( newpid ) {
    close( fds[0] );
    if( *from_addr ) {
      sprintf( fname,"From %s\n",from_addr );
      write( fds[1],fname,strlen( fname ) );
    }
    sprintf( fname,"X-NoNews: 1\n" );
    write( fds[1],fname,strlen( fname ) );
    while( ( numread = fread( fname, 1, N_BUFSIZE-1, f ) ) )
		write( fds[1], fname, numread );
    close( fds[1] );
/* Wait for the child to terminate else the child will
   become zombie */
    waitpid( newpid,NULL,0 );
  } else {		/* Child process exec's nlist */
    close( fds[1] );
    close(0);
    dup2( fds[0] , 0);	/* Pipe is connected to exec'd process stdin */
    close( fds[0] );
    sprintf( fname,"%s/nlist",PATH );
    if( execv( fname, command_array ) == -1 ) {
      Log(1,1, "nlistd: unable to execute %s\n",fname);
      fclose( f );
      exit( 0 );	/* child has to be killed explicitly if exec didn't do it */
    }
  }  
}

#ifndef DEBUG
/* nlistd's main function */
int main( int argc, char *argv[] ) {
  ARTICLEID	*a;
  int	childpid,fd;
  int	x,ec;
  char		fname[N_BUFSIZE];
  FILE		*f;
  time_t	t;
  struct tm	*tmp;
  unsigned long	old_lct;
  
  init_commonvars();
  if( x = setjmp( finish ) ) {
    if( config != NULL )
	fclose( config );	/* See remark at read_config */
    switch( x ) {
      case NO_LIST:
      	printf( "No config file for %s list\n",list_name );
      	break;
    }
    exit( x - 1 );
  }
/* grab the list name */
  if( argc < 2 ) {
    fprintf( stderr,"Usage: nlistd <listname>\n" );
    exit( 1 );
  }
  strcpy( list_name,argv[1] );
  read_config( list_name,0 );
  fclose( config );	/* Release config lock immediately */
  config = NULL;
  if( !nntp_newstolist ) {
    fprintf( stderr,"News-to-list message transfer is not enabled for %s - exiting\n",
    		list_name );
    exit( 1 );
  }
  if( *nntp_newsgroup == 0 ) {
    fprintf( stderr,"newsgroup not specified in config - exiting\n" );
    exit( 1 );
  }
  if( ( childpid = fork() ) < 0 )
	fprintf( stderr,"nlistd: Can't fork first child" );
  else
  if( childpid > 0 )
	exit( 0 );	/* Kill parent if 
			   the daemon was not started in the
			   background */
  setpgrp();		/* Becomes process group leader */
  if( ( fd = open( "/dev/tty",O_RDWR ) ) >= 0 ) {
    ioctl( fd,TIOCNOTTY,(char *)NULL );	/* Lose control tty */
    close( fd );
  }
  for( fd = 0 ; fd < NOFILE ; fd++ )
	close( fd );
  errno = 0;	/* Probably set by close */
  chdir( "/" );	/* So that the original dir can be unmounted */
  umask( nlist_umask );
  childpid = getpid();
  sprintf( fname,"%s/%s.nlistd_pid",PATH,list_name );
  if( ( f = fopen( fname,"w" ) ) == NULL ) {
    Log( 1,1,"nlistd: can't open: %s",fname );
    exit( 1 );
  }
  fprintf( f,"%i\n",childpid );
  fclose( f );
  Log( 2,1,"nlistd: Nlist daemon started" );
  
  read_status();
  t = (time_t)lastcheck_time;
  tmp = localtime( &t );
  switch( nntp_newart_access ) {
    case NEWART_STAT:
        Log( 3,1,"nlistd: using current_article_id: %li",current_article_id );
  	break;
  	
    case NEWART_NEWNEWS:
        Log( 3,1,"nlistd: using lastcheck_time %04u.%02u.%02u %02u:%02u:%02u",
  			tmp->tm_year+1900,tmp->tm_mon+1,tmp->tm_mday,
  			tmp->tm_hour,tmp->tm_min,tmp->tm_sec );
  	break;
  }
  article_list = NULL;
/* At last the update process */
  while( 1 ) {
    old_lct = lastcheck_time - nntp_msgwindow;
    if( old_lct < firstcheck_time )
    	old_lct = firstcheck_time;
    lastcheck_time = (unsigned long)time( NULL );
    process_new_news( old_lct );
    a = article_list;
    while( a != NULL ) {
      Log( 4,1,"Fetching article: %s\n",a->articleid );
      ec = fetch_news_article( a->articleid );
      if( ec ) {
        Log( 2,1,"Error %u fetching article %s\n",ec,a->articleid );
      } else {
        send_article_to_list();	
      }
      a = a->link;
    }
    write_status();
    if( *nntp_spooldir != 0 )
    	scan_spool_dir();
    sleep( nntp_newschecktime );
  }
/* not reached */
}

#endif

#ifdef DEBUG
/* Unit test */
int main( int argc, char *argv[] ) {
  ARTICLEID	*a;
  FILE		*f;
  int		ec,x;
  char		fname[N_BUFSIZE];

  init_commonvars();
  if( x = setjmp( finish ) ) {
    if( config != NULL )
	fclose( config );	/* See remark at read_config */
    switch( x ) {
      case NO_LIST:
      	printf( "No config file for %s list\n",list_name );
      	break;
    }
    exit( x - 1 );
  }
/* grab the list name */
  if( argc < 2 ) {
    fprintf( stderr,"Usage: nlistd <listname>\n" );
    exit( 1 );
  }
  strcpy( list_name,argv[1] );
  read_config( list_name,0 );
  fclose( config );	/* Release config lock immediately */
  config = NULL;
  read_status();
 /* approximately 1990 */
  lastcheck_time = 20*365*24*60*60L;
  printf( "nlistd: using lastcheck_time %lu\n",lastcheck_time );
  article_list = NULL;
  process_new_news( lastcheck_time );
  a = article_list;
  while( a != NULL ) {
    printf( "ID: %s\n",a->articleid );
    a = a->link;
  }
  a = article_list;
  while( a != NULL ) {
    printf( "Fetching article: %s\n",a->articleid );
    ec = fetch_news_article( a->articleid );
    if( ec ) {
      printf( "Error %u fetching article %s\n",ec,a->articleid );
    } else {
      sprintf( fname,"%s/%s.nlistd_tmp",PATH,list_name );
      if( ( f = fopen( fname,"r" ) ) == NULL ) {
        printf( "Cannot open: %s\n",fname );
      } else {
        if( *from_addr )
        	printf( "From %s\n",from_addr );
        while( fgets( fname,1000, f ) != NULL )
        	printf( "L: %s",fname );
        fclose( f );
        send_article_to_list();	
      }
    }
    a = a->link;
  }
}
#endif
