/************ nntp.c
	NNTP interface for nlist. Written by Gabor Paller
			    **********/
			    
#include	<stdio.h>
#include	<string.h>
#include	<sys/stat.h>
#include	<unistd.h>
#include	<fcntl.h>
#include	<sys/timeb.h>
#include	<time.h>
#include	<errno.h>
#include	<sys/types.h>
#include	<dirent.h>
#include	"config.h"
#define		N_EXTERN	extern
#include	"commonvars.h"
/* #define		DEBUG */
#if	defined	DEBUG
#define		write_smtp_line	debug_smtp_line
#endif

void scan_spool_dir();

static char	tempbuffer[1002];
static char	tempbuffer2[1002];
static char	sname[N_BUFSIZE];

static char *nntp_daynames[] = {
  "Sun","Mon","Tue","Wed","Thu","Fri","Sat"
};

static char *nntp_monthnames[] = {
  "Jan","Feb","Mar","Apr","May","Jun",
  "Jul","Aug","Sep","Oct","Nov","Dec"
};

void debug_smtp_line( char debug[] ) {
  printf( "NNTP: %s",debug );
}

/* Opens an NNTP session to the NNTP server. Returns 0 if OK, nonzero else */
int open_nntp_session() {
  int	retry = 10;

  strcpy( protstr,"NNTP" );
l0:
  if( retry == 0 ) {
    Log( 1,1,"NNTP error: %i:'%s'\n",errno,tempbuffer );
    return 1;
  }
  if( open_socket( nntp_server, nntp_port ) ) {
    Log( 1,1,"Couldn't open NNTP connection" );
    return 4;
  }
  do {
    socket_read_line( tempbuffer,1000 );
    if( strncmp( tempbuffer,"200",3 ) == 0 )	/* OK */
	break;
    if( strncmp( tempbuffer,"201",3 ) == 0 ) {	/* No posting allowed */ 
      Log( 1,1,"NNTP error: %s",tempbuffer );
      return 2;
    }
/* Some garbage arrived. Terminate the connection,
   close the socket and retry. I don't know how it can happen
   but it happens ... :-( */
    --retry;
    close_socket();
    sleep( 1 );		/* Wait 1 second */
    goto l0;
  } while( 1 );
/* This code is of little importance. I use INN on my testing
   machine and INND must be switched into reader mode if
   I post news articles from the server. In general case,
   this check is not needed */
  if( ( strstr( tempbuffer,"INN" ) != NULL ) &&
      ( strstr( tempbuffer,"NNRP" ) == NULL ) ) {
    socket_write_line( "mode reader\r\n" );
    socket_read_line( tempbuffer,1000 );
    if( strstr( tempbuffer,"NNRP" ) == NULL ) {
      Log( 1,1,"Cannot switch NNTP server to reader mode: %s",
		tempbuffer );
      return 3;
    }	
  }
  return 0;
}

/* Closes the NNTP session */
int close_nntp_session() {
  socket_write_line( "\n.\n" );
  socket_read_line( tempbuffer,1000 );
  if( strncmp( tempbuffer,"240",3 ) != 0 ) {
    Log( 1,1,"NNTP posting failed: %s",tempbuffer );
    return 1;
  }
  socket_write_line( "QUIT\n" );
  socket_read_line( tempbuffer,1000 );
  close_socket();
  return 0;  
}

/* Calculates a simple hash value to be used for message id */
int make_hash( FILE *infile ) {
  int hash = 0xABCD1234;
  char	*p,c;
  while( fgets( tempbuffer,1000,infile ) != NULL ) {
    p = tempbuffer;
    while( c = *(p++) ) {
      hash = hash << 1 | ( ( hash & 0x80000000 ) == 0 );
      hash ^= ((int)c) & 0xFF;
    } 
  }
  return hash;
}

/* Posts the letter to the newsgroup. Returns 0 if OK, error code else */
int nntp_post( char from[], char tname[] ) {
  FILE	*letter;
  
  if( ( letter = fopen( tname,"r" ) ) == NULL ) {
    Log( 1,1,"Cannot open letter file %s for NNTP posting",
                  tname );
    return 1;
  }
  return nntp_post_from_file( from,letter,TRUE );
}

int nntp_post_from_file( char from[], FILE *letter, bool notspooled ) {
  int	hash,len;
  char	hostname[128],domainname[128],message_id[128];
  struct timeb ts;
  struct tm *tp;
  long mimepos;
  bool	filter_mime,save_pos,subject_found;
  char	*p,c;
 
  fseek( letter,0L,SEEK_SET );
/* Calculate a simple hash for Message-ID */
  hash = make_hash( letter );
  fseek( letter,0L,SEEK_SET );

/* Opens nntp session, returns if error */
#if !defined DEBUG
  if( open_nntp_session() ) {
    fclose( letter );
    return 3;
  }
  socket_write_line( "POST\n" );
  socket_read_line( tempbuffer,1000 );
  if( strncmp( tempbuffer,"340",3 ) != 0 ) {
    Log( 1,1,"NNTP POST error: %s",tempbuffer );
    fclose( letter );
    return 2;
  }
#endif

/* From header goes out */
  sprintf( tempbuffer,"From: %s\n",
                subst_lname( nntp_from_address,sender_full_address,
		"news",
		nntp_list_address ) );
  write_smtp_line( tempbuffer );
/* Entry path */
  write_smtp_line( "Path: entry\n" );
/* Newsgroups */
  sprintf( tempbuffer,"Newsgroups: %s\n",nntp_newsgroup );
  write_smtp_line( tempbuffer );
/* Subject. First let's see if we have a non-empty subject string.
   Subject is mandatory according to NNTP spec */
  subject_found = FALSE;
  p = subject_string;
  while( *p && ( *p != '\n' ) ) {
    if( ( *p == ' ' ) || ( *p =='\t' ) )
    	++p;
    else {
      subject_found = TRUE;
      break;
    }
  }
  if( subject_found )
  	sprintf( tempbuffer,"Subject: %s\n",subject_string );
  else
	sprintf( tempbuffer,"Subject: %s no subject\n",list_title );
  write_smtp_line( tempbuffer );
/* Message ID */
  gethostname( hostname,127 );
  getdomainname( domainname,127 );
/* This is again my testing environment where domains
   are not properly set up. My Linux box returns (none)
   as domain name and it is not accepted by the news server.
   We fix it here. */
  if( ( *domainname == 0 ) || ( strncmp( domainname,"(none)",6 ) == 0 ) )
      strcpy( domainname,"localdomain" );
/* Prefetch time because we will need it for Message-ID as well */
  ftime( &ts );
  tp = gmtime( &(ts.time) );
/* Creating Message-ID header */
  sprintf( message_id,"<%08X.%u.%u@%s.%s>",hash,getpid(),tp->tm_sec,
  				hostname,domainname );
  sprintf( tempbuffer,"Message-ID: %s\n",message_id );
  write_smtp_line( tempbuffer );
/* Date. We don't rely on the date header of the original
   mailer of the letter as mail programs can often produce
   hideous date fields. The news post is timestamped by
   nlist */
  sprintf( tempbuffer,"Date: %s, %02i %s %04i %02i:%02i:%02i GMT\n",
              nntp_daynames[tp->tm_wday],tp->tm_mday,
	      nntp_monthnames[tp->tm_mon],tp->tm_year+1900,
	      tp->tm_hour,tp->tm_min,tp->tm_sec );
  write_smtp_line( tempbuffer );
/* Needed to avoid mail-news loops caused by nlistd */
  sprintf( tempbuffer,"X-Sender: nlist %s\n",NLIST_VERSION );
  write_smtp_line( tempbuffer );
  if( *nntp_replyto != 0 ) {
    sprintf( tempbuffer,"Reply-To: %s\n",subst_lname( nntp_replyto,
		sender_full_address,
		"news",
		nntp_list_address ) );
    write_smtp_line( tempbuffer );
  }
/* Now send out message headers that nlist doesn't handle.
   These may be MIME content type and content length
   messages that are necessary to handle attachments */
  do {
    if( fgets( tempbuffer, 1000, letter ) == NULL )
	      	break;
l0:
    if( *tempbuffer == '\n' )
	      	break;
    if( ( strncasecmp( tempbuffer,"From:", 5 ) == 0 ) ||
/* Boundary is not a real SMTP/NNTP header. It is used by the
   news spooling routine and is filtered out here */
        ( strncasecmp( tempbuffer,"Boundary:",9 ) == 0 ) || 
	( strncasecmp( tempbuffer,"To:", 3 ) == 0 ) ||
	( strncasecmp( tempbuffer,"Subject:",8 ) == 0 ) ||
	( strncasecmp( tempbuffer,"Reply-To:", 9 ) == 0 ) ||
	( strncasecmp( tempbuffer,"X-Sender:",9 ) == 0 ) ||
	( strncasecmp( tempbuffer,"Errors-To:",10) == 0 ) ||
	( strncasecmp( tempbuffer,"Date:",5 ) == 0 ) ||
	( strncasecmp( tempbuffer,"From",4 ) == 0 ) ||
	( strncasecmp( tempbuffer,"Sender:",7 ) == 0 ) ||
	( strncasecmp( tempbuffer,"Message-ID:",11 ) == 0 ) ||
	( strncasecmp( tempbuffer,"Return-Receipt-To:",18 ) == 0 ) ||
	( strncasecmp( tempbuffer,"Received:",9 ) == 0 ) ||
	( strncasecmp( tempbuffer,"NNTP-Posting-Host",17 ) == 0 ) ) {
/* Garbage header may take several lines. We drop the lines until
   we get another header line or end-of-header section mark (empty line) */
       do {
         if( fgets( tempbuffer, 1000, letter ) == NULL )
	      	goto l1;
         if( *tempbuffer == '\n' )
	      	goto l1;
         if( ( *tempbuffer != ' ' ) && ( *tempbuffer != '\t' ) )
         	goto l0;
       } while( TRUE );
     }
    write_smtp_line( tempbuffer );
  } while( TRUE );
l1:
  write_smtp_line( "\n" );
/* Now send out the message content */
  while( fgets( tempbuffer,1000,letter ) != NULL ) {
/* Binary attachments must be filtered out if the NNTP server doesn't allow
   them in the newsgroup */
    if( nntp_nobinary && ( *boundary != 0 ) && ( strncmp( tempbuffer,"--",2 ) == 0 ) &&
        ( strstr( tempbuffer,boundary ) != NULL ) ) {
/* Let's remember this position, we may have to return */
      p = strstr( tempbuffer,boundary ) + strlen( boundary );
      if( strncmp( p,"--",2 ) == 0 ) {	/* If MIME end trailer */
        write_smtp_line( tempbuffer );
        break;
      }
      mimepos = ftell( letter );
      filter_mime = FALSE;
      save_pos = -1;
      while( fgets( tempbuffer2,1000,letter ) != NULL ) {
        if( *tempbuffer2 == '\n' )
        	filter_mime = TRUE;
        if( !filter_mime &&
            ( strncasecmp( tempbuffer2,"Content-type:",13 ) == 0 ) &&
            ( strcasestr( tempbuffer2,"text/plain" ) != NULL ) ) {
/* It's a text/plain part, we won't skip it, let's return where we found
   the boundary string */
          fseek( letter,mimepos,SEEK_SET );
          goto nofilter;
        }
        if( ( save_pos >= 0 ) &&
            ( strncmp( tempbuffer2,"--",2 ) == 0 ) &&
            ( strstr( tempbuffer2,boundary ) != NULL ) ) {
          fseek( letter,save_pos,SEEK_SET );
          goto cont_outer;
        }
        save_pos = ftell( letter );
      }
      break;	/* Break the outer loop too */
    }
nofilter:
    write_smtp_line( tempbuffer );
cont_outer:
  }
  fclose( letter );
#if !defined DEBUG
  close_nntp_session();
#endif
  if( nntp_newstolist )
	CheckArticleFile( message_id );
  return 0;
}

/* Check the article file if the article with the given message ID is in the
   file. If it is, return with TRUE, if not, add it to the file and return FALSE.
   Meanwhile remove stale entries (which are older than 2*nntp_newschecktime) */
bool CheckArticleFile( char message_id[] ) {
  FILE	*artfile,*artfile_bak;
  LOCK_STRUCT	lt;
  bool retval = FALSE;
  unsigned long    current_time,limit_time,art_time;
  char  *w1,*w2,inpfile[128],updfile[128];
  int	i,err;

  lt.l_type = F_RDLCK;
  lt.l_whence = 0;
  lt.l_start = 0;
  lt.l_len = 0;
  current_time = (unsigned long)time( NULL );
  sprintf( inpfile,"%s/%s.newsart",PATH,list_name );
  if( ( artfile = fopen( inpfile,"r" ) ) == NULL ) {
    if( ( artfile = file_open( inpfile,"w" ) ) == NULL ) {
      Log( 1,1,"Cannot create news article ID file: %s",inpfile );
      return FALSE;
    }
    fprintf( artfile,"%s %lu\n",message_id,current_time );
    fclose( artfile );
    return FALSE;
  }
/* We try ten times to lock the file */
  for( i = 0 ; i < 10 ; ++i ) {
    err = fcntl( fileno( artfile ), F_SETLK, &lt );
    if( err >= 0 )
	    break;
    if( ( err < 0 ) && ( errno == EAGAIN ) )
	    sleep( 1 );
  }
  if( err < 0 ) {
    Log( 1,1, "Cannot lock news article ID file for 10 seconds" );
    fclose( artfile );
    return FALSE;
  }
/* Calculate the expiration time */
  limit_time = current_time - nntp_msgexpiration;
  sprintf( updfile,"%s/%s.newsart_bak",PATH,list_name );
  if( ( artfile_bak = file_open( updfile, "w" ) ) == NULL ) {
    Log( 1,1,"Cannot open news article backup file" );
    fclose( artfile );
    return FALSE; 
  }
  while( fgets( tempbuffer, 80, artfile ) != NULL ) {
    if( ( w1 = strtok( tempbuffer," " ) ) == NULL )
        continue;
    if( ( w2 = strtok( NULL," " ) ) == NULL )
        continue;
    if( sscanf( w2,"%lu",&art_time ) < 1 )
        continue;
    if( art_time < limit_time )
    	continue;
    if( strcmp( w1,message_id ) == 0 )
    	retval = TRUE;
    fprintf( artfile_bak,"%s %lu\n",w1,art_time );
  }
  if( !retval )
    fprintf( artfile_bak,"%s %lu\n",message_id,current_time );
  fclose( artfile_bak );
  fclose( artfile );
  unlink( inpfile );
  rename( updfile,inpfile );
  return retval;
}

/* Spools the article that could not be posted to the newsgroup */
void spool_article( char tname[] ) {
  long tstamp;
  char fname[N_BUFSIZE];
  FILE	*f,*letter;
  int	err;
  LOCK_STRUCT	lt;
  
  if( *nntp_spooldir == 0 )
  	return;
/* Let's generate the filename */
  do {
    tstamp = (long)time( NULL ) & 0xFFFFFF;
    sprintf( fname,"%s/%lu.tmp",nntp_spooldir,tstamp );
    if( ( f = fopen( fname,"a" ) ) == NULL )
    	continue;
    lt.l_type = F_WRLCK;
    lt.l_whence = 0;
    lt.l_start = 0;
    lt.l_len = 0;
    err = fcntl( fileno( f ), F_SETLK, &lt );
    if( ( err < 0 ) && ( errno == EAGAIN ) ) {	/* if it cannot be locked */
      fclose( f );
      continue;
    }
    if( ftell( f ) != 0L ) {	/* if it was not new file */
      fclose( f );
      continue;
    }
/* We have the unique filename now, let's start writing */
    break;
  } while( TRUE );
  fprintf( f,"From: %s\n",sender_full_address );
  fprintf( f,"Subject: %s\n",subject_string );
  fprintf( f,"Boundary: %s\n",boundary );
  if( ( letter = fopen( tname,"r" ) ) == NULL ) {
    Log( 1,1,"Cannot open letter file %s for NNTP spooling",
                  tname );
    return;
  }
/* Copy the letter into the spool file */
  while( fgets( tempbuffer2,1000,letter ) != NULL )
  	fputs( tempbuffer2,f );
  Log( 4,1,"Article spooled in file %s",fname );
  fclose( f );
}

/* Cuts down string at CR or LF */
void strip_crlf( char string[] ) {
  char c;
  
  while( c = *string ) {
    if( ( c == '\n' ) ||
        ( c == '\r' ) ) {
      *string = 0;
      return;
    }
    ++string;
  }
}

/* Sends spooled article. Gets the name of the spooled article file
   and sends it out using NNTP. Returns 0 if OK, error code else */
int send_spooled_article( char fname[] ) {
  FILE	*f;
  LOCK_STRUCT	lt;
  char		buf[N_BUFSIZE];
  int		err;
  
  if( ( f = fopen( fname,"r+" ) ) == NULL )
  	return 100;
  lt.l_type = F_WRLCK;
  lt.l_whence = 0;
  lt.l_start = 0;
  lt.l_len = 0;
  err = fcntl( fileno( f ), F_SETLK, &lt );
  if( ( err < 0 ) && ( errno == EAGAIN ) ) {	/* if it cannot be locked */
    fclose( f );
    return 101;
  }
/* 0 (OK) is returned if the file is corrupted because
   retry will not help here. Corrupted file is considered
   "done" and is deleted by the caller */
  if( fgets( buf,N_BUFSIZE-1,f ) == NULL ) {
corrupted:
    fclose( f );
    return 0;
  }
  strip_crlf( buf );
  if( strncmp( buf,"From: ",6 ) != 0 )
  	goto corrupted;
  strcpy( sname,buf+6 );
  sender_full_address = sname;
  if( fgets( buf,N_BUFSIZE-1,f ) == NULL )
  	goto corrupted;
  strip_crlf( buf );
  if( strncmp( buf,"Subject: ",9 ) != 0 )
  	goto corrupted;
  strcpy( subject_string,buf+9 );
  if( fgets( buf,N_BUFSIZE-1,f ) == NULL )
  	goto corrupted;
  strip_crlf( buf );
  if( strncmp( buf,"Boundary: ",10 ) != 0 )
  	goto corrupted;
  strcpy( boundary,buf+10 );
  Log( 4,1,"Article is being sent out from spool file %s",fname );
  return nntp_post_from_file( sender_full_address, f, FALSE );
}

/* Scans the directory for spooled articles */
void scan_spool_dir() {
  DIR	*dir;
  struct dirent *de;
  char	fname[ N_BUFSIZE ];
  struct stat sb;

  if( ( dir = opendir( nntp_spooldir ) ) == NULL )
  	return;
  while( ( de = readdir( dir ) ) != NULL ) {
    sprintf( fname,"%s/%s",nntp_spooldir,de->d_name );
    stat( fname,&sb );
    if( S_ISREG( sb.st_mode ) && !send_spooled_article( fname ) )
		unlink( fname );
  }
  closedir( dir ); 
}
